# Script to generate processed data from raw train announcements and train messages

library(data.table)
library(jsonlite)
library(lubridate)
library(stringr)

# process_announcements() generates
# /processed/announcements.csv - All announcements in one file
# --> with extra columns: Date, DateTime
# /processed/announcements_latest.csv - Only the latest announcement for each arrival/departure in one file
# --> with extra columns: Date, DateTime
# /processed/trains.csv - One row per train including arrival/departure in one file
# --> with extra columns: Date, DateTime, DateHour, Hour, DwellTime, Status, ArrivalDelay, DepartureDelay, Line, ShortTrain, RemainingMinutesToDepartureAtArrival, DepartureDelayAtArrival
process_announcements <- function(path, locationSignature) {
  sourcePath <- file.path(path, "stream", "TrainAnnouncement", locationSignature)
  targetPath <- file.path(path, "processed", locationSignature)
  
  #create target path if not exists
  dir.create(targetPath, recursive=TRUE, showWarnings = FALSE)
  
  #Look for raw files
  announcementFiles <- list.files(sourcePath, pattern=paste("TrainAnnouncement_", locationSignature, "_[0-9]+.csv",sep=""), full.names=TRUE)
  message(paste("Reading", length(announcementFiles), "files"))
  
  #announcements.csv
  read_raw_csv <- function (file) {
    d <- fread(file)
    if (!("TimeReceived" %in% names(d))) {
      d <- cbind(TimeReceived=as.POSIXct(NA),d)
    }
    d$EstimatedTimeAtLocation <- as.POSIXct(d$EstimatedTimeAtLocation)
    d
  }
  announcements <- rbindlist(lapply(announcementFiles, read_raw_csv), use.names=TRUE, fill=TRUE)

  announcements$Date <- as.IDate(coalesce(announcements$TimeAtLocation,announcements$TechnicalDateTime))
  announcements$DateTime <- coalesce(announcements$TimeAtLocation, announcements$TechnicalDateTime)
  
  announcements <- announcements[order(DateTime)]
  
  fwrite(announcements, file=file.path(targetPath, "announcements.csv"))
  
  #announcements_latest.csv
  announcements_latest <- announcements[Deleted==FALSE,tail(.SD, 1), by = ActivityId] #keep only latest non-deleted update
  fwrite(announcements_latest, file=file.path(targetPath, "announcements_latest.csv"))
  
  #trains.csv
  traininfo <- announcements_latest[,lapply(.SD, function(x) tail(x[!is.na(x)],1)), by=.(Date,TechnicalTrainIdent)] #keep last non-na value for each column
  
  trains <- announcements_latest[,.(PlannedArrival=.SD[ActivityType=="Ankomst", TechnicalDateTime],
                                     ActualArrival=.SD[ActivityType=="Ankomst", TimeAtLocationWithSeconds],
                                     PlannedDeparture=.SD[ActivityType=="Avgang", TechnicalDateTime],
                                     ActualDeparture=.SD[ActivityType=="Avgang", TimeAtLocationWithSeconds]
  ), by=.(Date,TechnicalTrainIdent)]
  
  trains <- merge(trains, traininfo, by=c("Date", "TechnicalTrainIdent"), all.x=TRUE)
  
  #add extra columns
  trains <- trains[, ':='(
    DateHour = as.POSIXct(floor_date(DateTime, "hour")),
    Weekday = wday(DateTime),
    Hour = hour(DateTime),
    DwellTime = difftime(ActualDeparture, ActualArrival, unit="mins"),
    ArrivalDelay = difftime(ActualArrival, PlannedArrival, unit="mins"),
    DepartureDelay = difftime(ActualDeparture, PlannedDeparture, unit="mins"),
    RemainingMinutesToDepartureAtArrival = pmax(0, difftime(PlannedDeparture, ActualArrival, unit="mins")),
    DepartureDelayAtArrival = pmax(0, difftime(ActualArrival, PlannedDeparture, unit="mins")),
    ShortTrain = grepl("ANA031", Deviation),
    Line = str_match(trains$ProductInformation, "\\{'Code': 'PNA09[0-9]', 'Description': '([0-9X]+)'\\}")[,2]
  )]
  
  trains <- trains[Canceled == TRUE, Status := "Canceled"]
  trains <- trains[ArrivalDelay >= 6 | DepartureDelay >=6, Status := "Delayed"]
  trains <- trains[(is.na(PlannedDeparture) | DepartureDelay < 6) 
                   & (is.na(PlannedArrival) | ArrivalDelay < 6), Status := "On time"]
  
  trains <- trains[order(DateTime)]
  
  fwrite(trains, file=file.path(targetPath, "trains.csv"))
}

# process_messages() generates
# /processed/messages.csv - All announcements in one file
# /processed/messages_latest.csv - Only the latest announcement for each arrival/departure in one file

process_messages <- function(path, locationSignature) {
  sourcePath <- file.path(path, "stream", "TrainMessage", locationSignature)
  targetPath <- file.path(path, "processed", locationSignature)
  
  #create target path if not exists
  dir.create(targetPath, recursive=TRUE, showWarnings = FALSE)
  
  #Look for raw files
  messageFiles <- list.files(sourcePath, pattern=paste("TrainMessage_", locationSignature, "_[0-9]+.csv",sep=""), full.names=TRUE)
  message(paste("Reading", length(messageFiles), "files"))

  #messages.csv
  read_raw_csv <- function (file) {
    d <- fread(file)
    if (!("TimeReceived" %in% names(d))) {
      d <- cbind(TimeReceived=as.POSIXct(NA),d)
    }
    d
  }
  messages <- rbindlist(lapply(messageFiles, read_raw_csv), use.names=TRUE, fill=TRUE)
  messages <- messages[order(StartDateTime)]
  fwrite(messages, file=file.path(targetPath, "messages.csv"))
  
  #messages_latest.csv
  messages_latest <- messages[,lapply(.SD, function(x) tail(x[!(is.na(x) | (as.character(x) == ""))],1)), by=.(EventId)] #keep last non empty value for each column
  messages_latest <- messages_latest[,EndDateTime:=coalesce(EndDateTime,PrognosticatedEndDateTimeTrafficImpact,LastUpdateDateTime)] #use prognosticated end date if no end date set
  fwrite(messages_latest, file=file.path(targetPath, "messages_latest.csv"))
}


process_tnk117_dataset <- function(path, locationSignature) {
  trains.tnk117 <- fread(file.path(dataPath, "processed", "Sub", "trains.csv"))
  #Filter
  trains.tnk117 <- trains.tnk117[Operator == "SLL" & !Canceled &!Deleted & DwellTime >= 0 & DwellTime < 5 & DepartureDelay < 15]
  trains.tnk117 <- trains.tnk117[,c(#"PlannedArrival", "ActualArrival", "PlannedDeparture", 
                                    "Line", "Weekday", "Hour", "Canceled", "Deviation",
                                    "FromLocation", "InformationOwner", "LocationSignature", "MobileWebLink",
                                    "Operator", "OtherInformation", "ProductInformation", "Service",
                                    "ToLocation", "TrackAtLocation", "TrainComposition", "TrainOwner",
                                    "TypeOfTraffic", "WebLink", "WebLinkName", "ViaFromLocation",
                                    "ViaToLocation", "ArrivalDelay", 
                                    "RemainingMinutesToDepartureAtArrival", "DepartureDelayAtArrival",
                                    "ShortTrain", "DepartureDelay")]
  fwrite(trains.tnk117, file=file.path(targetPath, "trains_tnk117.csv"))
  
}



#Example:

#process_announcements("/Users/nilbr50/OneDrive - Linköpings universitet/MUR/Data", "Sub")
#process_messages("/Users/nilbr50/OneDrive - Linköpings universitet/MUR/Data", "Sub")
#process_tnk117_dataset("/Users/nilbr50/OneDrive - Linköpings universitet/MUR/Data", "Sub")


